#include <iostream>
#include <windows.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "ip_icmp.h"
#include "iphdr.h"
#include "ip_opt.h"

#define MAX 65536

using namespace std;

unsigned short checksum(unsigned char *addr, int count);

int main(int argc, char *argv[])
{
  WORD wVersionRequested = MAKEWORD(2,2); 
  WSADATA wsaData;
  int size;
  hostent *host = NULL; 
  icmphdr icmp, *icmpRecv;
  iphdr *ipRecv;
  SOCKET sock;
  int lenght;
  sockaddr_in sendSockAddr;
  fd_set mySet;
  timeval tv;
  unsigned short int pid = GetCurrentProcessId();
  char buffer[MAX];
  char ipOptions[MAX_IPOPTLEN];
  bool recv = false;

  if ((argc > 9) || (argc < 3))
  {
    cerr << "Syntaxe:\n\t"
        << argv[0] << "seznam adres 2 - 9 adres" 
        << endl;
    return -1;
  }
  if (WSAStartup(wVersionRequested, &wsaData) != 0)
  {
    cerr << "Nelze inicializovat sokety" << endl;
    return -1;
  }
  // Zjistime informace z DNS o posledn adrese
  if ((host = gethostbyname(argv[argc - 1])) == NULL)
  {
    cerr << "Spatna clova adresa" << endl;
    WSACleanup();
    return -1;
  }
  // Vytvorime soket
  if ((sock = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)) == INVALID_SOCKET)
  {
    cerr << "Nelze vytvorit soket" << endl;
    WSACleanup();
    return -1;
  }
  // Pripravime strukturu pro sendto
  sendSockAddr.sin_family = AF_INET;
  sendSockAddr.sin_port = 0;
  memcpy(&(sendSockAddr.sin_addr), host->h_addr, host->h_length);
  // Zaplneni ICMP hlavicky
  icmp.type = ICMP_ECHO;
  icmp.code = 0;
  icmp.un.echo.id = pid;
  icmp.checksum = 0;
  icmp.un.echo.sequence = 1;
  icmp.checksum = checksum((unsigned char *)&icmp, sizeof(icmphdr));
  // Zaplneni volitelna casti IP hlavicky
  memset(ipOptions, 0, MAX_IPOPTLEN);
  ipOptions[IPOPT_OPTVAL] = IPOPT_LSRR;
  ipOptions[IPOPT_OLEN] = 3 + (argc - 1) * 4;
  ipOptions[IPOPT_OFFSET] = IPOPT_MINOFF;
  for (int i = 1; i < argc ; i++, ipOptions[IPOPT_OFFSET] += 4)
  {
    if ((host = gethostbyname(argv[i])) == NULL)
    {
      cerr << "Spatna mezi adresa" << endl;
      close(sock);
      WSACleanup();
      return -1;
    }
    memcpy((void *)&ipOptions[ipOptions[IPOPT_OFFSET] - 1], (void *)host->h_addr, host->h_length);
  }
  ipOptions[IPOPT_OFFSET]  = 4;
  if (setsockopt(sock, IPPROTO_IP, IP_OPTIONS, ipOptions, ipOptions[IPOPT_OLEN]) == -1)
  {
    cout << "Nelze nastavit explicitn smrovn" << endl;
    close(sock);
    WSACleanup();
    return -1;
  }
  if ((sendto(sock, (char *)&icmp, sizeof(icmphdr), 0, (sockaddr *)&sendSockAddr, sizeof(sockaddr)) == -1))
  {
    cerr << "Problem s odeslanim dat" << endl;
    close(sock);
    WSACleanup();
    return -1;
  }
  do
  {
    FD_ZERO(&mySet);
    FD_SET(sock, &mySet);
    tv.tv_sec = 5;
    tv.tv_usec = 0;
    if (select(sock + 1, &mySet, NULL, NULL, &tv) < 0)
    {
      cerr << "Selhal select" << endl;
      break;
    }
    if (FD_ISSET(sock, &mySet))
    {
      size = sizeof(sendSockAddr);
      if ((lenght = recvfrom(sock, buffer, MAX, 0, (sockaddr *)&sendSockAddr, &size)) == SOCKET_ERROR)
      {
            close(sock);
            WSACleanup();
            return -1;
      }
      cout << "Prijato " << lenght << " bytu ze " << inet_ntoa((in_addr)sendSockAddr.sin_addr) << endl;
      ipRecv = (iphdr *)buffer;
      icmpRecv = (icmphdr *) (buffer + ipRecv->ihl * 4);
      if (ipRecv->ihl * 4 + sizeof(icmphdr) > (unsigned int)lenght)
      {
            continue;
      }
      if ((icmpRecv->type == ICMP_ECHOREPLY) && (icmpRecv->code == 0) && (icmpRecv->un.echo.id == pid) && (icmpRecv->un.echo.sequence == 1))
      {
	      cout << "Jedna se o odpoved na mou zadost (neuveritelne)" << endl;
	      recv = true;
      }
      if ((icmpRecv->type == ICMP_DEST_UNREACH) && (icmpRecv->code == ICMP_SR_FAILED))
      {
            if ((unsigned int)lenght < ipRecv->ihl * 4 + 2 * sizeof(icmphdr) + sizeof(iphdr) + ipOptions[IPOPT_OLEN])
            {
                  ipRecv = (iphdr *)((char *)icmpRecv + sizeof(icmphdr));
                  icmpRecv = (icmphdr *)(((char *)ipRecv) + ipRecv->ihl * 4);
                  if ( (icmpRecv->type == ICMP_ECHO) && (icmpRecv->code == 0) 
                               && (icmpRecv->un.echo.id == pid) && (icmpRecv->un.echo.sequence == 1))
                  {
                          cout << "Je to ICMP paket signalizujici zahozeni naseho ECHO pozadavku" 
                               << " - explicitni smerovani selhalo" << endl;
                          recv = true;
                  }
            }
      }
    }
    else
    {
      cerr << "Nic neprislo" << endl;
      break;
    }
  } while (!recv);
  close(sock);
  WSACleanup();
  return 0;
}

/* Funkce checksum je opsana z RFC dokumentu. */
unsigned short checksum(unsigned char *addr, int count)
{
  register long sum = 0;
  
  while( count > 1 )  
  {
    /*  This is the inner loop */
    sum += * ((unsigned short *) addr)++;
    count -= 2;
  }  
  /*  Add left-over byte, if any */
  if( count > 0 )
    sum += * (unsigned char *) addr;
  /*  Fold 32-bit sum to 16 bits */
  while (sum>>16)
    sum = (sum & 0xffff) + (sum >> 16);
  return ~sum;
}

